;;;
;;; file:    position-stack.el
;;; version: 1.0
;;; date:    12 June 1994
;;;
;;; Barry Kaplan
;;; blk@orbit.dalmatian.com
;;; Copyright (c) 1994 The Dalmatian Group, Inc
;;;
;;; <place standard MIT copyright here> In other words, this
;;; is _free_ software. Do with it as you please.
;;;
;;;
;;; Description
;;; -----------
;;; This is a very simple package which provides a "position
;;; stack". That is, a stack of buffer/window/screen positions.
;;;
;;; The following capabilies are provided:
;;;
;;;  - Push the current position
;;;  - Pop and goto the position on the top of the stack
;;;  - Goto (but not pop) any position in the stack
;;;  - Remove the top position from the stack
;;;  - Clear the stack (ie, remove all positions from the stack)
;;;
;;; In addition a popup menu is provided (currently there is 
;;; no builtin support for adding an entry to the menubar). The
;;; menu contains:
;;;
;;;     Push Position
;;;     Pop Position
;;;     Remove Top
;;;     Clear Stack
;;;     -------------
;;;     <position>
;;;     ...
;;;
;;; Where <position>'s are menu items dynamically added to the end 
;;; of menu. Each entry contains the buffer name, the line number, 
;;; and the text on the line of the position. Selecting one of these 
;;; menu items will perform a "goto" to the position (ie, the position 
;;; will not be removed from the stack.)
;;;
;;;
;;;
;;; Installation
;;; ------------
;;; 1. Requires stack-f.el from elib (v0.6).
;;;
;;; 2. Requires func-menu (v2.11 or later).
;;;
;;; 3. Include the following in your .emacs:
;;;
;;;      (require 'position-stack)
;;;
;;; 4. Optionally bind public functions to key/mouse. I have in
;;;    my .emacs:
;;;
;;;      (define-key global-map [(meta button3)]        'pos-popup-menu)
;;;      (define-key global-map [(control kp_add)]      'pos-push-position)
;;;      (define-key global-map [(control kp_subtract)] 'pos-pop-position) 
;;;
;;; 5. Optionally set the values of customization variables. See the
;;;    "Customization" section below for details.
;;;
;;;
;;;------------------------------------------------------------
;;;
;;; Test cases:
;;;
;;; Set 1
;;; -----
;;; This set of test cases tries to prove that the position is not
;;; set correctly for states of the position buffer.
;;;
;;; 1.1  Buffer exists
;;; 1.2  Buffer does not exist
;;; 1.2.1  Buffer was associated a file
;;; 1.2.3  Buffer was not associated with a file
;;; 1.3  Buffer was killed and reopened
;;; 1.3.1  pos-prompt-for-reopen == t
;;; 1.3.1.1  Yes is selected
;;; 1.3.1.2  No is selected
;;; 1.3.1.3  Abort is selected. If the initiating event was a pop,
;;;          the position should not be removed from the stack
;;; 1.3.2  pos-prompt-for-reopen == nil
;;;
;;; Set 2
;;; -----
;;; This set of test cases tries to prove that the customization
;;; variable 'pos-hidden-buffer-selection-position' does not
;;; cause the the buffer to be displayed in the correct window.
;;;
;;; From a requirements view, these test cases should be crossed with
;;; with Set 1, but is not because the buffer selection is independent
;;; of the window/screen selection (also because this is free software
;;; and it not cost justified.)
;;;
;;; 2.1  pos-hidden-buffer-selection-position == 'current-window
;;; 2.2  pos-hidden-buffer-selection-position == 'another-window
;;; 2.3  pos-hidden-buffer-selection-position == 'new-screen
;;;
;;;------------------------------------------------------------


(require 'stack-f)    ; tested with elib v0.6
(require 'func-menu)  ; tested with v2.11


;; =======               ====================================================
;; =====   Customization   ==================================================
;; =======               ====================================================


(defvar pos-auto-position-popup fume-auto-position-popup
  "Set this to nil if you don't want the popup menu to always appear in the
corner of the window, in which case it will track the mouse position instead.
The default value is inherited from fume-auto-position-popup")

(defvar pos-max-submenu-items fume-max-items
  "Maximum number of elements in a position submenu. The default value
is inherited from fume-max-items.")

(defvar pos-hidden-buffer-selection-policy 'current-window
  "Defines where/how poped or goto'ed buffers are displayed. This variable
only applies when the buffer is NOT currently displayed in some winodow.
The options are:
     
    'current-window -- The buffer is placed in the current window.
    'another-window -- The buffer is placed in another window in the 
                       current screen.
    'new-screen     -- The buffer is placed in a newly created screen.")

(defvar pos-prompt-for-reopen t
  "Defines whether the user should be prompted before reopening a 
position's file whose buffer no longer exists. t means prompt. nil
means open without prompting.")

(defvar pos-format-position-menu-item 'position-format-menu-item
  "The function to use for formatting the position menu items. The default
is to display the buffer name, the line number, and the text on the position
line.")


(defun position-format-menu-item (position)
  "Returns a string to be used as the menu item for the given position."

  (format "%s (%d): %s" 
	  (position-buffer-name position)
	  (position-line-number position)
	  (position-context position))
)


;; =======                ===================================================
;; =====   Public Methods   =================================================
;; =======                ===================================================


(defun pos-push-position ()

  "Push point on the top of the position stack. If the current point
is the same as the position on the top of the stack, the position is
NOT pushed on the stack (and an error message is generated.)"

  (interactive)

  ;;
  ;; We only push the marker on the stack if it is not at the
  ;; same location as the current top marker.
  ;;
  (if (equal (position-marker (stack-top position-stack)) (point-marker))
      (error "Point is already on the top of the position stack.")
    (stack-push position-stack (position-construct))
    (message "Point pushed on position stack."))
)


(defun pos-pop-position ()

  "Set point to the position on the top of the position stack. 

  If the position is in a buffer which is displayed in another 
  window in the selected screen that window is selected. The position
  is removed from the the stack.

  If the position is in a buffer which is displayed in another
  window not in the selected screen, that screen is raised, and
  the window and screen are selected. (Of course, the window still
  has the final say as to which screen is selected.) The position
  is removed from the stack

  If the position is in a buffer which is not displayed in any window,
  the action taken depends on the value of pos-buffer-selection-policy. 
  (See variable documentation for a description of the various 
  behaviors.)  In any case the position is removed from the stack.

  If the postion's buffer no longer exists, the user is prompted
  if the file should be opened. Choosing Yes opens the file and removes
  the position from the stack. Choosing No simly removes the position
  from the stack. Choosing Abort/Quit does not open the file nor does
  it remove the position from the stack."

  (interactive)
  
  (if (stack-empty position-stack)
      (error "The position stack is empty."))

  (pos-goto-position (stack-pop position-stack) t)

  (message "Position popped to top of stack.")
)


(defun pos-remove-top ()

  "Removes the top position from the position stack."

  (interactive)
  (stack-pop position-stack)
)


(defun pos-clear-stack ()

  "Clears the position stack (ie, all positions are removed from the
stack."

  (interactive)
  (stack-clear position-stack)
  (message "Position stack cleared.")
)


(defun pos-popup-menu ()

  "Pops up a menu with the items:

   Push Position
   Pop Position
   Remove Top
   Clear Stack
   ----------
   <position>
   ...

Selecting a <position> item performs a 'goto position'. Goto position
is equivalent to a pos-pop-position except that the position is never
removed from the stack.

The function referenced by the variable pos-format-position-menu-item
can be set to a function which returns a string to be used as the
position menu item."

  (interactive)

  
  (popup-menu (append position-fixed-menu-items
		      (if (not (stack-empty position-stack))
			  (cons "----"
				(position-dynamic-menu-items))))) 
)



;; =======                 ==================================================
;; =====   Private Methods   ================================================
;; =======                 ==================================================


;;
;; The postion data structure is a vector containing:
;;
;;   0 -> marker
;;   1 -> buffer name
;;   1 -> file name (may be nil)
;;   2 -> line number of point
;;   3 -> the text on the line (for context in the menu)
;;

(defvar position-stack (stack-create)
  "The stack of markers. This is a private subobject.")

(defsubst position-construct ()
  "Constructs a position for the current point."
  (vector (point-marker)                           ; 0 marker
	  (count-lines 1 (point))                  ; 1 line number
	  (buffer-name)                            ; 2 buffer name
	  buffer-file-truename                     ; 3 file name 
	  (buffer-substring (save-excursion        ; 4 context
			      (beginning-of-line)
			      (point))
			    (save-excursion
			      (end-of-line)
			      (point))))
)

(defsubst position-reset-marker (position)
  "Resets the marker of the position to the current point."
  (aset position 0 (point-marker))
)


;;
;; Position data accessors.
;;

(defsubst position-marker (position)
  "Returns the position's marker."
  (elt position 0)
)

(defsubst position-line-number (position)
  "Returns the position's line number in the buffer."
  (elt position 1)
)

(defsubst position-buffer-name (position)
  "Returns the postion's buffer name."
  (elt position 2)
)

(defsubst position-file-name (position)
  "Returns the file name of the postion's buffer."
  (elt position 3)
)

(defsubst position-context (position)
  "Returns the text on the position line."
  (elt position 4)
)


;;
;; Values derived from the position data.
;;

(defsubst position-buffer-p (position)
  "Returns t if positions buffer currently exsist."
  (bufferp (marker-buffer (elt position 0)))
)

(defsubst position-buffer (position)
  "Returns the position's buffer."
  (and (position-buffer-p position)
       (marker-buffer (elt position 0)))
)

(defsubst position-point (position)
  (marker-position (position-marker position))
)

(defsubst position-window-p (position)
  "Returns t if the position is currently displayed in a buffer"
  (and (position-buffer-p position)
       (windowp (get-buffer-window (position-buffer position) t)))
)

(defsubst position-window (position)
  "Returns a window which is displaying the position's buffer."
  (and (position-window-p position)
       (get-buffer-window (position-buffer position) t))
)

(defsubst position-file-name-p (position)
  "Returns the file name of the postion's buffer."
  (stringp (elt position 3))
)



;; ------------------------------------------------------------



(defvar position-open-question "The buffer for file %s no longer exists.
Open the file? ")


(put 'position-error               'error-conditions '(error))
(put 'position-no-reopen           'error-conditions '(error position-error))
(put 'position-file-does-not-exist 'error-conditions '(error position-error))
(put 'position-no-assoc-file       'error-conditions '(error position-error))


(defun pos-goto-position (position repush)

  "Goto the specified marker.

TBD: Need to update/complete the doc."

  (condition-case err
      (let (the-buffer the-window the-screen)

	;;
	;; First we  get the position buffer. 
	;;
	(setq the-buffer 
	      (or 
	       ;;
	       ;; The buffer still exists.
	       ;;
	       (position-buffer position)
	       
	       ;;
	       ;; The buffer was killed and reopened since the position
	       ;; was pushed (ie, the position's marker's buffers is nil.)
	       ;;
	       (and (position-file-name-p position)
		    (get-file-buffer (position-file-name position)))
	       
	       ;;
	       ;; The buffer was not associated with a file.
	       ;;
	       (and (not (position-file-name-p position))
		    (signal 'position-no-assoc-file nil))
	       
	       ;;
	       ;; The buffer no longer exists; we prompt the user if we
	       ;; should reopen it (unless pos-prompt-for-reopen is nil.)
	       ;;
	       (let* ((reopen (or (not pos-prompt-for-reopen) ; we should just do the reopen
				  (y-or-n-p-maybe-dialog-box 
				   (format position-open-question 
					   (position-file-name position))))))
		 (if (not reopen)
		     (signal 'position-no-reopen nil))
	     
		 (if (not (file-exists-p (position-file-name position)))
		     (signal 'position-file-does-not-exist nil))

		 (find-file-noselect (position-file-name position)))

	       ;;
	       ;; TBD: This should be handled in the above case.
	       ;;
	       (error (concat "Could not open file " (position-file-name position)))))


    ;;
    ;; Ok, we got the buffer. Now get the window to display it in.
    ;;
    (cond
     ;;
     ;; The buffer IS currently displayed in a window. In this case we
     ;; always use the existing window.
     ;;
     ;; TBD: This could be a little cleaner.
     ;;
     ((setq the-window (get-buffer-window the-buffer t))
      (select-window the-window)
      (setq the-screen (selected-screen))
      (if (or (not (screen-visible-p the-screen))
	      (not (screen-totally-visible-p the-screen)))
	  (progn
	    (sit-for 0)
	    (make-screen-visible (window-screen the-window)))))
      
     ;;
     ;; The buffer is NOT displayed in a window, and the user
     ;; specified that we should use the current window to display the
     ;; buffer.
     ;;
     ((eq pos-hidden-buffer-selection-policy 'current-window)
      (set-window-buffer (selected-window) the-buffer))
     
     ;;
     ;; The buffer is NOT displayed in a window, and the user
     ;; specified that we should display the buffer in another window
     ;; in the currently selected screen.
     ;;
     ((eq pos-hidden-buffer-selection-policy 'another-window)
      (switch-to-buffer-other-window the-buffer))

     ;;
     ;; The buffer is NOT displayed in a window, and the user
     ;; specified that we should display the buffer in a newly
     ;; created screen.
     ;;
     ((eq pos-hidden-buffer-selection-policy 'new-screen)
      (switch-to-buffer-other-screen the-buffer))
     
     );end-cond
  

    ;;
    ;; Now we have to set the position.
    ;;
    (cond
     ;;
     ;; The position buffer existed. Thus we can use the marker to set
     ;; point.
     ;;
     ((position-buffer-p position)
      (set-window-point (selected-window) (position-point position)))

     ;;
     ;; The position buffer did not exist. Thus we need to use the line
     ;; number to set point. Also, we reset the marker if we are popping
     ;; the position.
     ;;
     (t
      (goto-line (position-line-number position))
      (and repush
	   (position-reset-marker position)))

     );end-cond

    );end-let

    ;;
    ;; TBD: Describe the error handling...
    ;;
    (t
     (cond
      ((eq (car err) 'quit)
       (and repush
	    (stack-push position-stack position))
       (error "Pop aborted."))
     
      ((eq (car err) 'position-no-reopen)
       (and (not repush)
	    (message "Positon removed from stack.")))
     
      ((eq (car err) 'position-file-does-not-exist)
       (and repush
	    (stack-push position-stack position)))

      ((eq (car err) 'position-no-assoc-file)
       (error "Position's buffer was killed and was not associated with a file."))

      (t
       (error "position-stack: UNHANDLED ERROR."))
     ))

    );end-condition-case

);end pos-goto-position



;; ----------------------------------------------------------------------



(defvar position-fixed-menu-items
  '("Position Stack"
    ["Push Position" pos-push-position t]
    ["Pop Position"  pos-pop-position (not (stack-empty position-stack))]
    ["Remove Top"    pos-remove-top   (not (stack-empty position-stack))]
    ["Clear Stack"   pos-clear-stack  (not (stack-empty position-stack))]
  )
)


(defsubst position-dynamic-menu-items ()

  "Returns a menu-item list of the positions currently on the stack.
The variable 'pos-format-position-menu-item' is used to generated
the menu-item string."

  (let* ((count 0)
	 (position-items
	  ;;
	  ;; This handles the list of lists returned by fume-split.
	  ;;
	  (mapcar
	   (function
	    (lambda (sublist)
	      (cons (format "Postions %d" (setq count (1+ count)))
		    ;;
		    ;; This handles the real list of markers.
		    ;;
		    (mapcar
		     (function
		      (lambda (position)
			(vector (funcall pos-format-position-menu-item position)
				(list 'pos-goto-position
				      position
				      nil)
				t)))
		     sublist))))
	   ;;
	   ;; Will return a list of lists. The nested lists will
	   ;; be the position markers. The size of each nested
	   ;; list will be no larger than `pos-max-submenu-items'.
	   ;;
	   (fume-split (stack-all position-stack) pos-max-submenu-items))))
    
    ;;
    ;; If we did NOT split the positions into submenus then we
    ;; need to move the items from the single submenu up to the
    ;; toplevel menu.
    ;;
    (if (= count 1)
	(cdr (car position-items))
      position-items))
)


;------------------------------

(provide 'position-stack)


